/* $Id$

This file is part of gldbg
Copyright © 2009  Mikko Rasa, Mikkosoft Productions
Distributed under the GPL
*/

#include <cstdlib>
#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <readline/readline.h>
#include <msp/core/except.h>
#include <msp/fs/dir.h>
#include <msp/io/print.h>
#include <msp/strings/lexicalcast.h>
#include "gldbg.h"
#include "glprint.h"

using namespace std;
using namespace Msp;

Application::RegApp<GlDbg> GlDbg::reg;

GlDbg::GlDbg(int argc, char **argv):
	cmd_interp(*this),
	process(vector<string>(argv+1, argv+argc)),
	flushing(false),
	got_sigchld(false)
{
	FS::Path libdir = FS::get_sys_lib_dir(argv[0], "gldbg");
	process.setenv("LD_PRELOAD", (libdir/"glwrap.so").str().c_str());
}

int GlDbg::main()
{
	catch_signal(SIGINT);
	catch_signal(SIGCHLD);
	set_loop_mode(TICK_BUSY);

	Application::main();
}

void GlDbg::launch()
{
	if(process.get_state()!=Process::INACTIVE)
		throw InvalidState("Program is already running");

	int fds[2];
	socketpair(AF_UNIX, SOCK_STREAM, 0, fds);
	sock_fd = fds[0];

	int flags = fcntl(sock_fd, F_GETFD);
	fcntl(sock_fd, F_SETFD, flags|FD_CLOEXEC);

	process.setenv("GLWRAP_FD", lexical_cast(fds[1]));
	process.launch();
	close(fds[1]);
}
	
void GlDbg::quit()
{
	if(process.get_state()!=Process::INACTIVE)
		throw InvalidState("Program is still running");
	exit(0);
}

void GlDbg::tick()
{
	if(got_sigchld)
	{
		got_sigchld = false;
		int ret = process.check();
		if(ret==0x100)
			IO::print("Target process exited normally\n");
		else if((ret>>8)==1)
			IO::print("Target process exited with status %d\n", ret&0xFF);
		else if((ret>>8)==2)
			IO::print("Target process terminated with signal %d\n", ret&0xFF);
		else if((ret>>8)==3)
		{
			IO::print("Target process stopped by signal %d\n", ret&0xFF);
			flushing = true;
		}
	}

	Process::State pstate = process.get_state();
	if((pstate!=Process::INACTIVE && pstate!=Process::STOPPED) || flushing)
		read_stream();
	else
	{
		char *line = readline("gldbg> ");
		if(line)
		{
			try
			{
				cmd_interp.execute(line);
			}
			catch(const Exception &e)
			{
				IO::print("%s\n", e.what());
			}
			free(line);
		}
		else if(pstate==Process::INACTIVE)
			exit(0);
	}
}

void GlDbg::read_stream()
{
	pollfd pfd = { sock_fd, POLLIN, 0 };
	int ret = poll(&pfd, 1, (flushing ? 0 : -1));
	if(ret>0)
	{
		char rbuf[1024];
		ret = read(sock_fd, rbuf, 1024);
		if(ret>0)
		{
			buffer.append(rbuf, ret);
			while(buffer.size()>buf_offset)
			{
				const char *data = buffer.data()+buf_offset;
				unsigned len = buffer.size()-buf_offset;
				int size = gldecoder_decode(0, data, len);
				if(size<0)
					break;
				tracer.decode(data, len);
				buf_offset += size;
			}
			if(buf_offset>8192)
			{
				buffer.erase(0, buf_offset);
				buf_offset=0;
			}
		}
	}
	else if(flushing)
		flushing = false;
}

void GlDbg::sighandler(int sig)
{
	if(sig==SIGCHLD)
		got_sigchld = true;
}
